from dataclasses import dataclass, field
from typing import Dict, cast
import betterproto
from pysmelt.smelt_telemetry.data import CommandEvent, CommandFinished, Event


@dataclass
class RetcodeTracker:
    """
    Simple subscriber that maps commands to their return codes
    """

    retcode_dict: Dict[str, int] = field(default_factory=dict)

    def process_message(self, message: Event):
        (variant, event_payload) = betterproto.which_one_of(message, "et")
        if variant == "command":
            event_payload = cast(CommandEvent, event_payload)
            (command_name, command_payload) = betterproto.which_one_of(
                event_payload, "CommandVariant"
            )

            if command_name == "finished":
                command_payload = cast(CommandFinished, command_payload)
                self.retcode_dict[event_payload.command_ref] = (
                    command_payload.out.status_code
                )
            else:
                pass
        else:
            pass

    def total_executed(self) -> int:
        return len(self.retcode_dict.items())

    def total_passed(self) -> int:
        return sum(1 for rc in self.retcode_dict.values() if rc == 0)

    def reset(self):
        self.retcode_dict = {}



from dataclasses import dataclass
from typing import Dict, Any


@dataclass
class RuntimeRequirements:
    num_cpus: int
    # This number is in MB
    max_memory_mb: int
    # timeout in seconds
    timeout: int
    # working_directory
    env: Dict[str, str]

    @classmethod
    def default(cls, env: Dict[str, str]):
        return cls(num_cpus=1, max_memory_mb=1024, timeout=600, env=env)

    @classmethod
    def from_dict(cls, indict: dict[str, Any]):
        num_cpus = indict["num_cpus"]
        max_memory_mb = indict["max_memory_mb"]
        timeout = indict["timeout"]
        env = indict["env"]

        return cls(
            num_cpus=num_cpus, max_memory_mb=max_memory_mb, timeout=timeout, env=env
        )



from pydantic.dataclasses import dataclass
from typing import ClassVar, Dict, Optional
from pathlib import Path
from pysmelt.path_utils import get_git_root
import toml
import os

from pprint import pprint


@dataclass(frozen=True)
class SmeltRC:

    smelt_root: str
    smelt_default_out: str
    smelt_rules_dir: str
    jobs: int

    @classmethod
    def default(cls):
        default_jobs = 8
        try:
            smelt_root = get_git_root()
        except:
            smelt_root = os.getcwd()

        return cls(
            smelt_root=smelt_root,
            smelt_default_out="smelt-out",
            smelt_rules_dir="smelt_rules",
            jobs=default_jobs,
        )

    @classmethod
    def try_load(cls):
        git_root = get_git_root()
        rc_path = Path(f"{git_root}/.smeltrc")

        if not rc_path.exists():
            return SmeltRC.default()

        stream = rc_path.read_text()

        try:
            rc_content = toml.loads(stream)
            return cls(
                smelt_root=rc_content["smelt_root"],
                smelt_default_out=rc_content["smelt_default_out"],
                smelt_rules_dir=rc_content["smelt_rules_dir"],
                jobs=rc_content["jobs"],
            )
        except toml.TomlDecodeError as exc:
            print(exc)
            raise RuntimeError(exc)

    @staticmethod
    def init_rc():
        default = SmeltRC.default()
        git_root = get_git_root()
        rc_path = Path(f"{git_root}/.smeltrc")
        with open(rc_path, "w") as outfile:
            toml.dump(default.__dict__, outfile)
        smelt_rules_dir = Path(f"{git_root}/{default.smelt_rules_dir}").mkdir(exist_ok=True)
        pprint(f"Initialized .smeltrc at {rc_path}")

    @property
    def abs_rules_dir(self) -> Path:
        git_root = get_git_root()
        return Path(f"{git_root}/{self.smelt_rules_dir}")


class SmeltRcHolder:
    _current_rc: ClassVar[SmeltRC] = SmeltRC.default()

    @staticmethod
    def current_rc() -> SmeltRC:
        return SmeltRcHolder._current_rc



from enum import Enum
from dataclasses import dataclass
from pathlib import Path


class SmeltPathType(Enum):
    Absolute = 1
    SmeltRootRelative = 2
    SmeltCommandDefRelative = 3


@dataclass
class SmeltPath:
    path_type: SmeltPathType
    path: str

    @property
    def as_str(self):
        return self.path

    @classmethod
    def abs_path(cls, path: str):
        return cls(path_type=SmeltPathType.Absolute, path=path)

    @classmethod
    def from_str(cls, path: str):
        if Path(path).is_absolute():
            path_type = SmeltPathType.Absolute
        else:
            path_type = SmeltPathType.SmeltRootRelative
        return cls(path_type=path_type, path=path)

    def __str__(self):
        return self.path

    def to_abs_path(self, smelt_root: str):
        if self.path_type == SmeltPathType.SmeltRootRelative:
            return f"${smelt_root}/{self.path}"
        elif self.path_type == SmeltPathType.Absolute:
            return f"{self.path}"
        else:
            raise NotImplementedError(f"Unhandled variant {self.path_type}")


@dataclass
class SmeltFilePath:
    path_type: SmeltPathType
    path: str

    @classmethod
    def abs_path(cls, path: str):
        return cls(path_type=SmeltPathType.Absolute, path=path)

    @classmethod
    def from_str(cls, path: str):
        if Path(path).is_absolute():
            path_type = SmeltPathType.Absolute
        else:
            path_type = SmeltPathType.SmeltCommandDefRelative
        return cls(path_type=path_type, path=path)

    def __str__(self):
        return self.path

    def to_abs_path(self, abs_cmd_path: str):
        if self.path_type == SmeltPathType.SmeltRootRelative:
            return f"${abs_cmd_path}/{self.path}"
        elif self.path_type == SmeltPathType.Absolute:
            return f"{self.path}"
        else:
            raise NotImplementedError(f"Unhandled variant {self.path_type}")



from typing_extensions import Annotated
from pathlib import Path
import typer
import yaml
from pysmelt.output import smelt_console
from pysmelt.interfaces import SmeltTargetType
from pysmelt.smelt_muncher import parse_smelt
from pysmelt.output_utils import pretty_print_tests
from pysmelt.pygraph import create_graph
from pysmelt.serde import SafeDataclassDumper
from typing import Optional
from typer import Exit

app = typer.Typer()


TlPath = Annotated[
    Path,
    typer.Argument(
        ...,
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
]

CommandPath = Annotated[
    Path,
    typer.Argument(
        exists=False,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
]


RulePath = Annotated[
    Path,
    typer.Argument(
        ...,
        exists=False,
        file_okay=True,
        dir_okay=True,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
]


def validate_type(value: str):
    if value not in SmeltTargetType._value2member_map_:
        raise Exit(
            f'Invalid value for "--tt". Possible values are {", ".join(SmeltTargetType._value2member_map_.keys())}'  # type: ignore
        )
    return value


@app.command(
    help="Lowers an smelt file to a command file",
)
def lower(
    smelt_file: TlPath,
    output: CommandPath = Path("command.yaml"),
):
    typer.echo(f"Validating: {smelt_file}")

    _, commands = parse_smelt(test_list=str(smelt_file))
    yaml.dump(commands, open(output, "w"), Dumper=SafeDataclassDumper, sort_keys=False)


@app.command(
    help="Executes an smelt file",
)
def execute(
    smelt_file: TlPath,
    tt: str = typer.Option("test", help="SMELT target type", callback=validate_type),
    target_name: Optional[str] = typer.Option(
        None, help="Target name -- if not provided, runs all the tests"
    ),
    rerun: bool = typer.Option(
        False, help="Rerun the commands that failed", is_flag=True
    ),
):

    graph = create_graph(str(smelt_file))
    if target_name:
        graph.run_one_test_interactive(target_name)
    else:
        graph.run_all_tests(tt)
    if rerun:
        graph.rerun()


@app.command(
    help="Executes an smelt file",
)
def validate(
    smelt_file: TlPath,
    tt: str = typer.Option("test", help="SMELT target type", callback=validate_type),
    target_name: Optional[str] = typer.Option(
        None, help="Target name -- if not provided, runs all the tests"
    ),
    rerun: bool = typer.Option(
        False, help="Rerun the commands that failed", is_flag=True
    ),
):

    graph = create_graph(str(smelt_file))
    smelt_console.print(f"[green] {smelt_file.name} is valid")
    pretty_print_tests(graph)


def main():
    app()


if __name__ == "__main__":
    app()



from typing import Dict, Generator, List, Optional, Tuple, cast

import betterproto
from pysmelt.interfaces import Command, SmeltTargetType, Target
from dataclasses import dataclass
from pysmelt.interfaces.paths import SmeltPath
from pysmelt.smelt_muncher import parse_smelt
from pysmelt.path_utils import relatavize_inp_path
from pysmelt.pysmelt import PyController, PyEventStream
from pysmelt.rc import SmeltRcHolder
from pysmelt.rerun import DerivedTarget, RerunCallback
import yaml
import time



from pysmelt.smelt_telemetry.data import Event
from pysmelt.smelt_client.commands import CfgDocker, CfgLocal, ConfigureSmelt


from pysmelt.subscribers.error_handler import SmeltErrorHandler
from pysmelt.subscribers.output_collector import OutputConsole
from pysmelt.subscribers.retcode import RetcodeTracker

from copy import deepcopy

from pysmelt.subscribers.stdout import StdoutPrinter, StdoutSink



def default_cfg(cmd_path: SmeltPath) -> ConfigureSmelt:
    rv = ConfigureSmelt()
    rc = SmeltRcHolder.current_rc()

    rv.job_slots = rc.jobs
    rv.smelt_root = rc.smelt_root
    rv.command_def_path = cmd_path.as_str
    rv.local = CfgLocal()
    

    return rv


def default_target_rerun_callback(
    target: Target, return_code: int
) -> Tuple[Target, bool]:
    """
    First pass at re-run logic -- currently we just rerun all tests that are tagged as tests

    Power users could supply their own logic, but we should define something that is robust and sane
    """

    requires_rerun = target.rule_type() == SmeltTargetType.Test and return_code != 0
    new_target = deepcopy(target)
    new_target.name = f"{new_target.name}_rerun"
    new_target.injected_state = {"debug": "True"}
    return (new_target, requires_rerun)


def maybe_get_message(
    listener: PyEventStream, blocking: bool = False
) -> Optional[Event]:
    try: 
        if blocking:
            message = listener.pop_message_blocking()
            event = Event.FromString(message)
        else:
            message = listener.nonblocking_pop()
            if message is None:
                return None
            event = Event.FromString(message)
        
        return event
    except RuntimeError as e: 
        """
        In expectation, we only return an error if the channel has been closed  

        We don't have any paths above this function to handle that error, so we'll handle it here as getting the message failing 
        """
        
        print(e)
        return None


def spin_for_message(
    listener: PyEventStream, backoff: float = 0.2, time_out: int = 10):
    """
    Utility for spinning for a message 
    """
    time_passed = 0 
    while time_passed < time_out:
        message = maybe_get_message(listener, blocking=False)
        if message:
            return message
        else:
            time.sleep(backoff) 
            time_passed += backoff

    raise RuntimeError(f"Before timeout of {time_out} expired!")



@dataclass
class PyGraph:
    """
    PyGraph is the python wrapper for the smelt runtime.
    """

    smelt_targets: Optional[Dict[str, Target]]
    """ 
    holds the original smelt targets (the python rules) that the user supplied. 

    This is used to re-generate new commands on failure
    """
    commands: List[Command]
    """ 
    holds all of the commands that are live in the graph -- some of these may not map back to an smelt target
    """
    controller: PyController
    """
    Handle to rust -- calls the function defined in the pyo3 bindings
    """
    
    retcode_tracker : RetcodeTracker

    def runloop(self, listener: PyEventStream):
        errhandler = SmeltErrorHandler()
        with OutputConsole() as console:
            while not listener.is_done():
                # tbh, this could be async
                # but async interopt with rust is kind of experimental and i dont want to do it yet
                message = maybe_get_message(listener, blocking=False)
                if message:
                    self.retcode_tracker.process_message(message)
                    console.process_message(message)
                    errhandler.process_message(message)
                if not message:
                    # add a little bit of backoff
                    time.sleep(0.01)

    def console_runloop(
            self, test_name: str, listener: PyEventStream, sink: StdoutSink
    ) -> Generator[bool, None, None]:
        """
        Generator that will try to consume as many `Event` messages as possible

        All of the stdout from the "test_name" command will be given to the `sink`.
        By default, sink should just be print

        The yielded value will be
        """
        stdout_tracker = StdoutPrinter(test_name, sink)
        errhandler = SmeltErrorHandler()
        while True:
            # tbh, this could be async
            # but async interopt with rust is kind of experimental and i dont want to do it yet
            message = maybe_get_message(listener, blocking=False)
            if message:
                self.retcode_tracker.process_message(message)
                stdout_tracker.process_message(message)
                errhandler.process_message(message)
            if not message:
                # add a little bit of backoff
                yield listener.is_done()

    def reset(self):
        pass
       # self.retcode_tracker.reset()
        

    def run_one_test_interactive(self, name: str, sink: StdoutSink = print):
        """
        Runs a single test, with a "sink" handle to process all of the stdout for that specific command

        By default, this will just print stdout + stderr to the screen -- it looks like you're running the command interactively 
        """
        self.reset()
        listener = self.controller.run_one_test(name)
        for is_done in self.console_runloop(name, listener, sink):
            if is_done:
                return
            time.sleep(0.1)

    def run_specific_commands(self, commands: List[Command]):
        self.reset()
        test_names = [command.name for command in commands]
        listener = self.controller.run_many_tests(test_names)
        self.runloop(listener)

    def run_all_tests(self, maybe_type: str):
        self.reset()
        listener = self.controller.run_all_tests(maybe_type)
        self.runloop(listener)

    

    def rerun(
        self,
        rerun_callback: RerunCallback = default_target_rerun_callback,
    ):
        if self.smelt_targets:
            """
            We create a dictionary that maps all of the target names to a "DerivedTarget", that holds all of the state for prospective targets that may need to be created 

            This dictionary maps the original target name -> DerivedTarget bundle
            """
            all_derived = {
                target: DerivedTarget.from_cb(
                    orig_target, rerun_callback(self.smelt_targets[target], retcode)
                )
                for target, retcode in self.retcode_tracker.retcode_dict.items()
                if (target in self.smelt_targets and (orig_target := self.smelt_targets[target]))
            }



            """
            For each target, we go through and see if any of the "derived" targets have "changed" from the original target 

            a target changing can mean one of three things

            1. it needs to be rerun 
            2. the command contents has changed and it does not need to be re-run (for example, when a if we want to have a debug build
            3. A dependency of the target has changed

            If a derived target has changed from its original, we lower it to a command, correct its dependencies and it is returned as part of the new commands list
            """

            new_commands = [
                (new_target, target.requires_rerun)
                for target in all_derived.values()
                if (new_target := target.get_new_command(all_derived)) is not None
            ]
            if not any(new_commands):
                return "No new commands were produced"
            all_commands_to_add, requires_rerun = map(list, zip(*new_commands))
            all_commands_to_add = cast(List[Command], all_commands_to_add)
            requires_rerun = cast(List[bool], requires_rerun)
            
            if not any(requires_rerun):
                print("No commands need a rerun is required!")
                return

            # TODO: its likely that we will also need to handle regenerating dependencies
            #       for a first pass functionality, lets ignore this for now

            self.add_commands(all_commands_to_add)
            commands_to_run = [
                command
                for command, rerun in new_commands 
                if rerun 
            ]
            self.run_specific_commands(commands_to_run)
        else:
            print(
                "Warning! Cannot auto re-run because no smelt targets have been provided"
            )

    def add_commands(self, commands: List[Command]):
        self.commands += commands
        commands_as_str = yaml.safe_dump([command.to_dict() for command in commands])
        self.controller.set_graph(commands_as_str)

    @classmethod
    def init(cls, smelt_targets: Dict[str, Target], commands: List[Command],  cfg : ConfigureSmelt ):
        cfg_bytes = bytes(cfg)
        graph = PyController(cfg_bytes)
        
        rv = cls(
            smelt_targets=smelt_targets, commands=[], controller=graph, retcode_tracker= RetcodeTracker()
        )
        rv.add_commands(commands)
        return rv

    # This is a testing utility
    @classmethod
    def init_commands_only(cls, commands: List[Command]):
        cfg = default_cfg(SmeltPath.from_str("."))
        return cls.init({}, commands, cfg)

def _create_cfg(smelt_test_list: str) -> ConfigureSmelt:
    command_def_path = relatavize_inp_path(SmeltRcHolder.current_rc().smelt_root,smelt_test_list)
    cfg = default_cfg(command_def_path)
    return cfg


def create_graph(smelt_test_list: str) -> PyGraph:
    cfg = _create_cfg(smelt_test_list)
    targets, command_list = parse_smelt(smelt_test_list)
    rv = PyGraph.init(targets, command_list, cfg)
    return rv

def create_graph_with_docker(smelt_test_list: str, docker_img: str) -> PyGraph:
    cfg = _create_cfg(smelt_test_list)
    cfg.docker = CfgDocker()
    cfg.docker.image_name = docker_img
    cfg.docker.additional_mounts = {}
    targets, command_list = parse_smelt(smelt_test_list)
    rv = PyGraph.init(targets, command_list, cfg)
    return rv                                        



from dataclasses import dataclass, field
from pysmelt.interfaces import Target, SmeltFilePath, SmeltTargetType, TargetRef
from typing import List, Dict


@dataclass
class raw_bash(Target):
    """
    Simple target for embedding raw bash commands in Smelt

    Environment variables avaible are:
        * ${GIT_ROOT}: the git root of the current git workspace
        * ${SMELT_ROOT}: the root of the smelt-workspace -- by default, this will be ${GIT_ROOT}/smelt
        * ${TARGET_ROOT}: the working space of the current target
    """

    cmds: List[str] = field(default_factory=list)
    debug_cmds: List[str] = field(default_factory=list)
    deps: List[TargetRef] = field(default_factory=list)

    def gen_script(self) -> List[str]:
        if "debug" in self.injected_state and self.debug_cmds:
            return self.debug_cmds
        else:
            return self.cmds

    def get_dependencies(self) -> List[TargetRef]:
        return self.deps


@dataclass
class raw_bash_build(raw_bash):
    @staticmethod
    def rule_type() -> SmeltTargetType:
        return SmeltTargetType.Build



from dataclasses import dataclass, field, asdict
from abc import ABC
from enum import Enum
from typing import Any, List, Dict, Literal
from pysmelt.interfaces.runtime import RuntimeRequirements
from pysmelt.interfaces.paths import SmeltFilePath


class SmeltTargetType(Enum):
    Test = "test"
    Stimulus = "stimulus"
    Build = "build"


TargetRef = str


@dataclass
class Target(ABC):
    """
    A target is a structure that holds logic to generate a `Command`

    Targets are higher level abstraction to generate a new command based off of certain new input criterea -- for instance, if a target fails, and you create a new command
    """

    name: str
    injected_state: Dict[str, Any] = field(init=False)

    def __post_init__(self):
        self.injected_state = {}

    def get_outputs(self) -> Dict[str, SmeltFilePath]:
        return {}

    def gen_script(self) -> List[str]:
        raise NotImplementedError

    @staticmethod
    def rule_type() -> SmeltTargetType:
        return SmeltTargetType.Test

    def runtime_env_vars(self) -> Dict[str, str]:
        return {}

    def runtime_requirements(self) -> RuntimeRequirements:
        return RuntimeRequirements.default(self.runtime_env_vars())

    def get_dependencies(self) -> List[TargetRef]:
        """
        Returns the targets that this target depends on
        """
        return []

    def get_dependent_files(self) -> List[str]:
        """
        Returns the files that this target depends on

        If any of these files change across invocations, the target will be re-executed
        """
        return []
